{% extends "base.html" %}

{% block main %}
{% include "_title.html" %}

<div class="box">
  <h2 class="subtitle">Providers</h2>

  {% for provider in providers %}
  <a class="button is-medium is-fullwidth" href="{{ provider.url }}" target="_blank">
    <figure class="icon">
      <img class="icon" src="{{ provider.logo }}" />
    </figure>
    <span>{{ provider.label }}</span>
  </a>
  {% endfor %}
</div>
{% endblock %}
