{% extends "security/base.html" %}
{% from "security/_macros.html" import render_field_with_errors, render_field, header %}

{% block title %}
Meltano - Change password
{% endblock %}

{% block content %}
{% include "security/_messages.html" %}
{{ header(_('Change password')) }}
<form action="{{ url_for_security('change_password') }}" method="POST" name="change_password_form">
  {{ change_password_form.hidden_tag() }}
  {{ render_field_with_errors(change_password_form.password) }}
  {{ render_field_with_errors(change_password_form.new_password) }}
  {{ render_field_with_errors(change_password_form.new_password_confirm) }}
  {{ change_password_form.submit(class_="button is-link") }}
</form>
{% endblock %}
